"""
You can run the .bat file on Linux with this hack:

cat convert_everything.bat | dos2unix | sed -e "s/python/python3/g" | bash

"""

import numpy as np
from scipy.linalg import orthogonal_procrustes
import os
import struct
import sys
import matplotlib.pyplot as plt
import scipy.signal
import argparse

WIDTH = 1920
HEIGHT = 1080
PROJECTION_PLANE_WIDTH = 4.9486 # in mm
FOCAL_LENGTH = 4 # in mm
FRAMERATE = 30.003444

parser = argparse.ArgumentParser()
parser.add_argument('inpath')
parser.add_argument('--start_frame', type=int, default=0)
parser.add_argument('--end_frame', type=int, default=-1)
parser.add_argument('--highpass_wavelength', type=float, default=1.25)
parser.add_argument('--bin')
parser.add_argument('--csv')
parser.add_argument('--npy')
parser.add_argument('--gui', action='store_true')

args = parser.parse_args()

inpath = args.inpath

filename = os.path.basename(inpath)

X = np.load(inpath)
print('X shape', X.shape)

if args.gui:
    plt.figure()
    plt.title('Point paths')
    plt.plot(X[0, :, 0], X[0, :, 1], label='Point #0')
    plt.plot(X[1, :, 0], X[1, :, 1], label='Point #1')
    plt.plot(X[2, :, 0], X[2, :, 1], label='Point #2')
    plt.legend()

# Input is in screen coordinates. Mirror the y-axis so that trigonometric functions work as usual.
X[:, :, 1] = HEIGHT - X[:, :, 1]

# Now X matrix y-axis points upwards.

def highpass(x, critical_wavelength_in_frames):
    cutoff = (1./critical_wavelength_in_frames)
    sos = scipy.signal.butter(2, cutoff, 'hp', fs=FRAMERATE, output='sos')
    x_sos = scipy.signal.sosfilt(sos, x)
    return x_sos

"""
X_before_filter = X.copy()

cutoff = FRAMERATE * args.highpass_wavelength
for coord in range(X.shape[2]):
    X[:, :, coord] -= X[:, :, coord].mean()

    for point in range(X.shape[0]):

        if args.highpass_wavelength > 0:
            X[point, :, coord] = highpass(X[point, :, coord], cutoff)


plt.figure()
plt.plot(X[0, :, 0], label='X filtered')
plt.plot(X_before_filter[0, :, 0], label='X raw')
plt.legend()
plt.show()
"""

crop_start = args.start_frame
crop_end = args.end_frame
X = X[:, crop_start:crop_end, :]

print('X shape after trimming', X.shape)

N_points, num_frames, dims = X.shape

ref = num_frames // 2

def estimate_angle(A, B):
    R_raw, sca = orthogonal_procrustes(A, B)
    R = R_raw.T

    # print(f"R[0][0] = {R[0][0]}")
    return 0.5 * (np.arccos(min(1.0, R[0][0])) + np.arcsin(R[1][0])), R, sca

def rotate(theta):
    return np.array([
        [np.cos(theta), -np.sin(theta)],
        [np.sin(theta), np.cos(theta)]])

def normalize(A):
    return (A - np.mean(A, 0)) / np.linalg.norm(A)

# Remove DC bias from input data
shift = X.mean(axis=0).mean(axis=0).reshape(1,1,2)
X -= shift

centroids = X.mean(axis=0)            # centroid position of each frame
Xc = X - centroids[np.newaxis, ...]   # local space with centroid as origin
# translation relative to reference point
# translation = centroids - centroids[ref:ref+1, :]
translation = centroids

show_frame = 300
ref_frame = ref

refpose = normalize(Xc[:, ref, :])

newpose = normalize(Xc[:, show_frame, :])
theta, R, sca = estimate_angle(refpose, newpose)

print(theta / (2*np.pi) * 360., 'degrees', sca)


def plot_triangle(row, label=''):
    plt.plot(row[[0,1,2,0], 0], row[[0,1,2,0], 1], label='shape '+label)
    #plt.scatter(row[0, 0], row[0, 1], label='Point #0' + label)
    #plt.scatter(row[1, 0], row[1, 1], label='Point #1' + label)
    #plt.scatter(row[2, 0], row[2, 1], label='Point #2' + label)

def plot_frame(X, show_frame, label=''):
    plot_triangle(X[:, show_frame, :], label=label)

if args.gui:
    plt.figure()
    plt.title(f"Points of frame {show_frame}")
    plot_frame(X, show_frame)
    plt.scatter(centroids[show_frame, 0], centroids[show_frame, 1], label='Centroid')
    plt.legend()

if args.gui:
    rotated_Xc_frame = Xc[:, show_frame, :] @ R
    rotated_ref_frame = Xc[:, ref_frame, :] @ R.T

    plt.figure()
    plt.title(f"Centered points of frame {show_frame}")
    plot_frame(Xc, show_frame, label=f" (Frame {show_frame})")
    plot_frame(Xc, ref_frame, label=f" (Reference frame {ref_frame})")
    plot_triangle(rotated_Xc_frame, label=f" (Rotated frame {show_frame})")
    plot_triangle(rotated_ref_frame, label=f" (Rotated ref {ref_frame})")
    plt.legend()
    # plt.show()

    plt.figure()
    plt.title('Centered point paths')
    plt.plot(Xc[0, :, 0], Xc[0, :, 1], label='Point #0')
    plt.plot(Xc[1, :, 0], Xc[1, :, 1], label='Point #1')
    plt.plot(Xc[2, :, 0], Xc[2, :, 1], label='Point #2')
    plt.plot(centroids[:, 0], centroids[:, 1], label='Centroid')
    plt.legend()
    # plt.show()

# Estimate rotation of each frame

rotation = np.zeros(num_frames)
directions = np.zeros((num_frames, 2))

A = normalize(Xc[:, ref, :])

for frame in range(num_frames):
    B = normalize(Xc[:, frame, :])
    theta, R, sca = estimate_angle(A, B)
    rotation[frame] = theta
    x, y = translation[frame]
    # print(f"[{frame}] ({x:.3f}, {y:.3f}) theta={theta:.6f}, sca={sca}")

# Workaround: reuse earlier theta for NaN values.
last_theta = 0.0
# print(f"rotation.shape {rotation.shape}")
for i in range(rotation.shape[0]):
    theta = rotation[i]
    #print(i, theta)
    if np.isnan(theta):
        print(i, rotation[i], theta, rotation.shape)
        rotation[i] = last_theta
        print(f"Warning: theta was NaN on frame {i}")
    else:
        last_theta = theta

directions[:,0] = np.cos(rotation)
directions[:,1] = np.sin(rotation)

rp = Xc[:, ref, :]
mean_error = 0.


for idx in range(num_frames):
    original = X[:, idx, :]
    recon = centroids[ref:ref+1] + translation[idx] + (rp @ rotate(rotation[idx]).T)
    error = np.linalg.norm(original-recon)
    assert not np.isnan(error)
    mean_error += error

mean_error /= num_frames

print(f"Mean reconstruction error: {mean_error} pixels")


def crossfade_loop(x, length):
    fade = np.linspace(0, 1, length, endpoint=False)

    x_faded = x.copy()[length:]
    x_faded[-length:] *= 1 - fade
    x_faded[-length:] += fade * x[:length]
    return x_faded

fade_length = int(FRAMERATE)
translation_looped = np.vstack([
    crossfade_loop(translation[:, 0], fade_length),
    crossfade_loop(translation[:, 1], fade_length),
]).transpose()

directions_looped = np.vstack([
    crossfade_loop(directions[:, 0], fade_length),
    crossfade_loop(directions[:, 1], fade_length),
]).transpose()

# Recompute the roll angles with tan(x) = sin(x)/cos(x) now that 'direction' has been looped.
nonzero = np.abs(directions_looped[:, 0]) > 1e-6
rotations_looped = np.zeros(directions_looped.shape[0])
rotations_looped[nonzero] = directions_looped[nonzero, 1] / directions_looped[nonzero, 0]

if args.gui:
    plt.figure()
    plt.title("Looping test")
    # plt.plot(translation_looped[0])
    plt.plot(np.tile(translation[fade_length:, 0], 2), label='X original', linestyle='--')
    plt.plot(np.tile(translation_looped[:, 0], 2), label='X')
    plt.plot(np.tile(translation_looped[:, 1], 2), label='Y')
    plt.vlines(translation_looped.shape[0], -200, 200, color='red', label='Loop point', alpha=0.3)
    plt.legend()
    #plt.plot()

plt.show()

def preprocess(x, channel_name):
    """Crossfades and high-pass filters the given signal 'x'."""

    fade_length = int(FRAMERATE)

    x_faded = crossfade_loop(x, fade_length)

    x_repeat = np.tile(x_faded, 3)

    critical_wavelength_in_frames = FRAMERATE * args.highpass_wavelength
    cutoff = (1./critical_wavelength_in_frames)
    # b, a = scipy.signal.butter(1, cutoff, analog=True, btype='high')
    sos = scipy.signal.butter(2, cutoff, 'hp', fs=FRAMERATE, output='sos')
    y_sos = scipy.signal.sosfilt(sos, x_repeat)
    y_sos_crop = y_sos[x_faded.shape[0]:(2*x_faded.shape[0])]
    y_sos_crop = crossfade_loop(y_sos_crop, fade_length)
    y_sos_crop_repeat = np.tile(y_sos_crop, 3)

    if args.gui and channel_name == 'x':
        plt.figure()
        plt.plot(x)
        plt.plot(x_repeat)

        plt.figure()

        plt.plot(x_repeat, 'r', label='Cross faded input signal')
        plt.plot(y_sos, 'k', label='High-pass filtered signal')
        plt.hlines(0, 0, len(x_repeat))
        plt.legend()
        plt.title('Signal debiasing')

        plt.figure()
        plt.title("Looping test")
        plt.plot(y_sos_crop_repeat)

        plt.show()
    
    return y_sos_crop

"""
translation = np.vstack([
    preprocess(translation[:, 0], 'x'),
    preprocess(translation[:, 1], 'y'),
]).transpose()
"""

"""
mm_per_pixel = PROJECTION_PLANE_WIDTH/1920
h_rads = np.arctan((translation[:, 0:1] * mm_per_pixel) / FOCAL_LENGTH)
v_rads = np.arctan((translation[:, 1:2] * mm_per_pixel) / FOCAL_LENGTH)

if args.gui:
    fig,ax=plt.subplots(1,1)
    ax.plot(v_rads, label="pitch")
    ax.plot(h_rads, label="yaw")
    ax.set_xlabel("frame")
    ax.set_ylabel("delta (radians)")
    plt.legend()
    plt.tight_layout()
    plt.show()
"""

if args.gui:
    fig,ax=plt.subplots(1,1)
    ax.plot(translation[:,0], label="x delta")
    ax.plot(translation[:,1], label="y delta")
    ax.set_xlabel("frame")
    ax.set_ylabel("delta (radians)")
    plt.legend()
    plt.tight_layout()
    plt.show()


def write(deltas, rot, dirs):
    N = deltas.shape[0]
    assert rot.shape[0] == N
    assert dirs.shape[0] == N

    if args.bin:
        outpath = args.bin
        print("Saving to", outpath)
        with open(outpath, 'wb') as f:
            f.write(struct.pack('4sI', 'SHAK'.encode('ASCII'), 1))
            f.write(struct.pack('IIIff', N, WIDTH, HEIGHT, FOCAL_LENGTH, FRAMERATE))
            for i in range(N):
                row = struct.pack('fffff', deltas[i][0], deltas[i][1], rot[i], dirs[i][0], dirs[i][1])
                f.write(row)

            print(f"Wrote {N} frames to {outpath}")

    if args.csv:
        with open(args.csv, 'wb') as f:
            for i in range(N):
                f.write(f"{deltas[i][0]},{deltas[i][1]},{rot[i]},{dirs[i][0]}, {dirs[i][1]}\n".encode('utf-8'))
        print(f"Also wrote {args.csv}")

    if args.npy:
        with open(args.npy, 'wb') as f:
            packed = np.hstack([deltas, rot.reshape(-1, 1)])
            print(f"Saving an .npy array of shape {packed.shape} to {args.npy}")
            np.save(f, packed)

write(translation_looped, rotations_looped, directions_looped)
