"""
Packs multiple exported tracks into a single PNG texture file.
The tracks must be exported with numpy.save and have the same shape.
That shape is (N, 3) where N is the same for each input.

This script also outputs a GLSL shader to decode the texture file.
Save this shader to a file with the '--glsl' switch.
"""

import math
import numpy as np
import argparse
from PIL import Image

parser = argparse.ArgumentParser()
parser.add_argument('--tracks', nargs='+')
parser.add_argument('--png')
parser.add_argument('--glsl')
args = parser.parse_args()

XDELTA = 0
YDELTA = 1
ROLL = 2

dim_names = ['XDELTA', 'YDELTA', 'ROLL']

print(args)

data = [np.load(path) for path in args.tracks]
for path, matrix in zip(args.tracks, data):
    assert matrix.shape == data[0].shape, f"{path} shape {matrix.shape} wasn't {data[0].shape}"

X = np.zeros((len(data), matrix.shape[0], matrix.shape[1]))
print(f"X.shape: {X.shape}")

for i, matrix in enumerate(data):
    X[i, :, :] = matrix

def quantize_simple(X):
    mins = []
    maxs = []

    for d in range(X.shape[2]):
        mins.append(X[:, :, d].min())
        maxs.append(X[:, :, d].max())

    print(f"xmin: {mins[XDELTA]}\nxmax: {maxs[XDELTA]}\nymin: {mins[YDELTA]}\nymax: {maxs[YDELTA]}")

    Xq = np.zeros(X.shape, dtype=np.uint8)

    for d in range(X.shape[2]):
        Xq[:, :, d] = np.round(255.0 * np.clip((X[:, :, d] - mins[d]) / (maxs[d]- mins[d]), 0.0, 1.0))

    return Xq, (mins, maxs)


def decode_simple(Xq, ranges):
    mins, maxs = ranges

    Xdec = np.zeros(Xq.shape)
    for d in range(Xdec.shape[2]):
        Xdec[:, :, d] = (((Xq[:, :, d] / 255.0) * (maxs[d] - mins[d])) + mins[d])

    return Xdec

Xq, ranges = quantize_simple(X)
Xdec = decode_simple(Xq, ranges)

def print_error(X, Xdec):
    diff = X - Xdec
    max_x_error = np.abs(diff[:,:,XDELTA]).max()
    max_y_error = np.abs(diff[:,:,YDELTA]).max()
    max_roll_error = np.abs(diff[:,:,ROLL]).max()
    mean_x_error = np.abs(diff[:,:,XDELTA]).mean()
    mean_y_error = np.abs(diff[:,:,YDELTA]).mean()
    mean_roll_error = np.abs(diff[:,:,ROLL]).mean()

    print(f"max_x_error {max_x_error:3f} pixels")
    print(f"max_y_error {max_y_error:3f} pixels")
    print(f"max_roll_error {max_roll_error/(np.pi*2)/360.} degrees")
    print(f"mean_x_error {mean_x_error:3f} pixels")
    print(f"mean_y_error {mean_y_error:3f} pixels")
    print(f"mean_roll_error {mean_roll_error/(np.pi*2)/360.} degrees")

print("\nQuantization errors")
print_error(X, Xdec)

if args.png:
    img = Image.fromarray(Xq)
    img.save(args.png)
    print(f"Saved {args.png}")

def make_shader(ranges):
    mins, maxs = ranges
    rangex = maxs[XDELTA]-mins[XDELTA]
    rangey = maxs[YDELTA]-mins[YDELTA]
    rangeroll = maxs[ROLL]-mins[ROLL]

    shader = f"""const int CAMSHAKE_STILL = 0;
const int CAMSHAKE_ONE_HAND = 1;
const int CAMSHAKE_SLOW_WALK = 2;
const int CAMSHAKE_FAST_WALK = 3;

vec4 readShakeTrack(int track, float time)
{{
    vec4 shake = texture2D(iShakeTexture, vec2((time * 30.) / 1024., (float(track) + 0.5) / 4.));
    shake *= vec4({rangex}, {rangey}, {rangeroll}, 1.0);
    shake += vec4({mins[XDELTA]}, {mins[YDELTA]}, {mins[ROLL]}, 0.);
    return shake;
}}

void applyShake(inout vec3 dir, vec4 shake)
{{
    mat2 rollMatrix = mat2(cos(shake.z), sin(shake.z), -sin(shake.z), cos(shake.z));
    dir.xy = rollMatrix * dir.xy;
    dir.xy += shake.xy / 1920.;
    dir = normalize(dir);
}}
"""

    return shader

glsl = make_shader(ranges)

if args.glsl:
    with open(args.glsl, 'w', encoding='utf-8') as f:
        f.write(glsl)
else:
    print(glsl)


